@@ -13,7 +13,9 @@ pub use std::borrow::Cow;
 
 pub use reqwest::Method;
 
+pub use crate::api::BodyError;
 pub use crate::api::Client;
 pub use crate::api::Endpoint;
+pub use crate::api::FormParams;
 pub use crate::api::Pageable;
-pub use crate::api::Pairs;
+pub use crate::api::QueryParams;
