@@ -4,6 +4,8 @@
 // option. This file may not be copied, modified, or distributed
 // except according to those terms.
 
+use reqwest::header;
+
 use crate::api::{ApiError, Client, Endpoint, Query};
 
 /// A query modifier that ignores the data returned from an endpoint.
@@ -26,11 +28,14 @@ where
 {
     fn query(&self, client: &C) -> Result<(), ApiError<C::Error>> {
         let mut url = client.rest_endpoint(&self.endpoint.endpoint())?;
-        self.endpoint.add_parameters(url.query_pairs_mut());
+        self.endpoint.parameters().add_to_url(&mut url);
 
-        let req = client
-            .build_rest(self.endpoint.method(), url)
-            .form(&self.endpoint.form_data());
+        let req = client.build_rest(self.endpoint.method(), url);
+        let req = if let Some((mime, data)) = self.endpoint.body()? {
+            req.header(header::CONTENT_TYPE, mime).body(data)
+        } else {
+            req
+        };
         let rsp = client.rest(req)?;
         if !rsp.status().is_success() {
             let v = serde_json::from_reader(rsp)?;
