@@ -4,14 +4,15 @@
 // option. This file may not be copied, modified, or distributed
 // except according to those terms.
 
-use std::collections::HashSet;
+use std::collections::BTreeSet;
 
 use chrono::{DateTime, Utc};
 use derive_builder::Builder;
 use itertools::Itertools;
 
-use crate::api::common::{self, NameOrId, SortOrder};
+use crate::api::common::{NameOrId, SortOrder};
 use crate::api::endpoint_prelude::*;
+use crate::api::ParamValue;
 
 /// Filters for issue states.
 #[derive(Debug, Clone, Copy, PartialEq, Eq)]
@@ -31,11 +32,17 @@ impl IssueState {
     }
 }
 
+impl ParamValue<'static> for IssueState {
+    fn as_value(self) -> Cow<'static, str> {
+        self.as_str().into()
+    }
+}
+
 #[derive(Debug, Clone)]
 enum Labels<'a> {
     Any,
     None,
-    AllOf(HashSet<Cow<'a, str>>),
+    AllOf(BTreeSet<Cow<'a, str>>),
 }
 
 impl<'a> Labels<'a> {
@@ -48,6 +55,12 @@ impl<'a> Labels<'a> {
     }
 }
 
+impl<'a, 'b: 'a> ParamValue<'static> for &'b Labels<'a> {
+    fn as_value(self) -> Cow<'static, str> {
+        self.as_str()
+    }
+}
+
 /// Filter issues by a scope.
 #[derive(Debug, Clone, Copy, PartialEq, Eq)]
 pub enum IssueScope {
@@ -69,28 +82,34 @@ impl IssueScope {
     }
 }
 
+impl ParamValue<'static> for IssueScope {
+    fn as_value(self) -> Cow<'static, str> {
+        self.as_str().into()
+    }
+}
+
 #[derive(Debug, Clone)]
 enum Assignee<'a> {
     Assigned,
     Unassigned,
     Id(u64),
-    Usernames(HashSet<Cow<'a, str>>),
+    Usernames(BTreeSet<Cow<'a, str>>),
 }
 
 impl<'a> Assignee<'a> {
-    fn add_query(&self, pairs: &mut Pairs) {
+    fn add_params<'b>(&'b self, params: &mut QueryParams<'b>) {
         match self {
             Assignee::Assigned => {
-                pairs.append_pair("assignee_id", "Any");
+                params.push("assignee_id", "Any");
             },
             Assignee::Unassigned => {
-                pairs.append_pair("assignee_id", "None");
+                params.push("assignee_id", "None");
             },
             Assignee::Id(id) => {
-                pairs.append_pair("assignee_id", &format!("{}", id));
+                params.push("assignee_id", *id);
             },
             Assignee::Usernames(usernames) => {
-                pairs.extend_pairs(usernames.iter().map(|value| ("assignee_username[]", value)));
+                params.extend(usernames.iter().map(|value| ("assignee_username[]", value)));
             },
         }
     }
@@ -113,6 +132,12 @@ impl<'a> ReactionEmoji<'a> {
     }
 }
 
+impl<'a, 'b: 'a> ParamValue<'a> for &'b ReactionEmoji<'a> {
+    fn as_value(self) -> Cow<'a, str> {
+        self.as_str().into()
+    }
+}
+
 /// Filter issues by weight.
 #[derive(Debug, Clone, Copy, PartialEq, Eq)]
 pub enum IssueWeight {
@@ -134,6 +159,12 @@ impl IssueWeight {
     }
 }
 
+impl ParamValue<'static> for IssueWeight {
+    fn as_value(self) -> Cow<'static, str> {
+        self.as_str()
+    }
+}
+
 /// Keys issue results may be ordered by.
 #[derive(Debug, Clone, Copy, PartialEq, Eq)]
 pub enum IssueOrderBy {
@@ -181,6 +212,12 @@ impl IssueOrderBy {
     }
 }
 
+impl ParamValue<'static> for IssueOrderBy {
+    fn as_value(self) -> Cow<'static, str> {
+        self.as_str().into()
+    }
+}
+
 /// Query for issues within a project.
 ///
 /// TODO: Negation (not) filters are not yet supported.
@@ -193,7 +230,7 @@ pub struct Issues<'a> {
 
     /// Filter issues with specific internal IDs.
     #[builder(setter(name = "_iids"), default, private)]
-    iids: HashSet<u64>,
+    iids: BTreeSet<u64>,
     /// Filter issues based on state.
     #[builder(default)]
     state: Option<IssueState>,
@@ -261,7 +298,7 @@ impl<'a> Issues<'a> {
 impl<'a> IssuesBuilder<'a> {
     /// Return an issue with an internal ID.
     pub fn iid(&mut self, iid: u64) -> &mut Self {
-        self.iids.get_or_insert_with(HashSet::new).insert(iid);
+        self.iids.get_or_insert_with(BTreeSet::new).insert(iid);
         self
     }
 
@@ -270,7 +307,7 @@ impl<'a> IssuesBuilder<'a> {
     where
         I: Iterator<Item = u64>,
     {
-        self.iids.get_or_insert_with(HashSet::new).extend(iter);
+        self.iids.get_or_insert_with(BTreeSet::new).extend(iter);
         self
     }
 
@@ -296,7 +333,7 @@ impl<'a> IssuesBuilder<'a> {
             set.insert(label);
             set
         } else {
-            let mut set = HashSet::new();
+            let mut set = BTreeSet::new();
             set.insert(label);
             set
         };
@@ -349,7 +386,7 @@ impl<'a> IssuesBuilder<'a> {
             set.insert(assignee);
             set
         } else {
-            let mut set = HashSet::new();
+            let mut set = BTreeSet::new();
             set.insert(assignee);
             set
         };
@@ -405,75 +442,42 @@ impl<'a> Endpoint for Issues<'a> {
         format!("projects/{}/issues", self.project).into()
     }
 
-    fn add_parameters(&self, mut pairs: Pairs) {
-        pairs.extend_pairs(
-            self.iids
-                .iter()
-                .map(|value| ("iids[]", format!("{}", value))),
-        );
-        self.state
-            .map(|value| pairs.append_pair("state", value.as_str()));
-        self.labels
-            .as_ref()
-            .map(|value| pairs.append_pair("labels", &value.as_str()));
-        self.with_labels_details
-            .map(|value| pairs.append_pair("with_labels_details", common::bool_str(value)));
-        self.milestone
-            .as_ref()
-            .map(|value| pairs.append_pair("milestone", value));
-        self.scope
-            .map(|value| pairs.append_pair("scope", value.as_str()));
+    fn parameters(&self) -> QueryParams {
+        let mut params = QueryParams::default();
+
+        params
+            .extend(self.iids.iter().map(|&value| ("iids[]", value)))
+            .push_opt("state", self.state)
+            .push_opt("labels", self.labels.as_ref())
+            .push_opt("with_labels_details", self.with_labels_details)
+            .push_opt("milestone", self.milestone.as_ref())
+            .push_opt("scope", self.scope)
+            .push_opt("my_reaction_emoji", self.my_reaction_emoji.as_ref())
+            .push_opt("weight", self.weight)
+            .push_opt("search", self.search.as_ref())
+            .push_opt("created_after", self.created_after)
+            .push_opt("created_before", self.created_before)
+            .push_opt("updated_after", self.updated_after)
+            .push_opt("updated_before", self.updated_before)
+            .push_opt("confidential", self.confidential)
+            .push_opt("order_by", self.order_by)
+            .push_opt("sort", self.sort);
+
         if let Some(author) = self.author.as_ref() {
             match author {
                 NameOrId::Name(name) => {
-                    pairs.append_pair("author_username", name);
+                    params.push("author_username", name);
                 },
                 NameOrId::Id(id) => {
-                    pairs.append_pair("author_id", &format!("{}", id));
+                    params.push("author_id", *id);
                 },
             }
         }
         if let Some(assignee) = self.assignee.as_ref() {
-            assignee.add_query(&mut pairs);
+            assignee.add_params(&mut params);
         }
-        self.my_reaction_emoji
-            .as_ref()
-            .map(|value| pairs.append_pair("my_reaction_emoji", value.as_str()));
-        self.weight
-            .map(|value| pairs.append_pair("weight", &value.as_str()));
-        self.search
-            .as_ref()
-            .map(|value| pairs.append_pair("search", value));
-        self.created_after.map(|value| {
-            pairs.append_pair(
-                "created_after",
-                &value.to_rfc3339_opts(chrono::SecondsFormat::Secs, true),
-            )
-        });
-        self.created_before.map(|value| {
-            pairs.append_pair(
-                "created_before",
-                &value.to_rfc3339_opts(chrono::SecondsFormat::Secs, true),
-            )
-        });
-        self.updated_after.map(|value| {
-            pairs.append_pair(
-                "updated_after",
-                &value.to_rfc3339_opts(chrono::SecondsFormat::Secs, true),
-            )
-        });
-        self.updated_before.map(|value| {
-            pairs.append_pair(
-                "updated_before",
-                &value.to_rfc3339_opts(chrono::SecondsFormat::Secs, true),
-            )
-        });
-        self.confidential
-            .map(|value| pairs.append_pair("confidential", common::bool_str(value)));
-        self.order_by
-            .map(|value| pairs.append_pair("order_by", value.as_str()));
-        self.sort
-            .map(|value| pairs.append_pair("sort", value.as_str()));
+
+        params
     }
 }
 
@@ -481,7 +485,7 @@ impl<'a> Pageable for Issues<'a> {}
 
 #[cfg(test)]
 mod tests {
-    use std::collections::HashSet;
+    use std::collections::BTreeSet;
 
     use crate::api::projects::issues::{IssueOrderBy, IssueScope, IssueState, IssueWeight, Issues};
 
@@ -502,12 +506,12 @@ mod tests {
     #[test]
     fn issue_labels_as_str() {
         let one_user = {
-            let mut set = HashSet::new();
+            let mut set = BTreeSet::new();
             set.insert("one".into());
             set
         };
         let two_users = {
-            let mut set = HashSet::new();
+            let mut set = BTreeSet::new();
             set.insert("one".into());
             set.insert("two".into());
             set
@@ -517,17 +521,12 @@ mod tests {
             (Labels::Any, "Any"),
             (Labels::None, "None"),
             (Labels::AllOf(one_user), "one"),
+            (Labels::AllOf(two_users), "one,two"),
         ];
 
         for (i, s) in items {
             assert_eq!(i.as_str(), *s);
         }
-
-        let many_labels = Labels::AllOf(two_users);
-        assert!(
-            many_labels.as_str() == "one,two" || many_labels.as_str() == "two,one",
-            "many_labels.as_str() did not join labels properly",
-        );
     }
 
     #[test]
@@ -569,6 +568,11 @@ mod tests {
         }
     }
 
+    #[test]
+    fn issue_order_by_default() {
+        assert_eq!(IssueOrderBy::default(), IssueOrderBy::CreatedAt);
+    }
+
     #[test]
     fn issue_order_by_as_str() {
         let items = &[
